#!/usr/bin/perl -w
# $Id$
#
# tv.cgi
#
#	CGI script to wrap TV program guide sites
#
#	Copyright (c)2005 Hiro Sugawara - All rights reserved
#	Permitted to re-distibute and modify under terms of GPL 2.0
#
# This script is a wrapper for TV program guide sites that offer iEPG
# file downloading. The script detects iEPG document download and
# redirects the document to an appropriate processing script without
# interaction with the user.
#
# This is a GET mehode taking two argumnets `u' and `__6ga_form_action__.'
#	u: the URL to show
#	__6ga_form_action__: wrapper for the form action in original documnet.
#
# The document at the given URL is retrieved by an internal version of
# wget.
# If the document is an HTML document, all non-graphic URL references
# are rewitten to be referrenced by this script before sent to the
# client.
# If the document is an iEPG file, it is transferred to the recording
# scheduler script (6ga.php).
#

use strict;
use Socket;
use IO::Handle;
use URI::Escape;
use File::Basename;

# Read in global configuration file
my $mydir	= dirname($0);
my $myname	= basename($0);
my $configfile	= $ENV{CONFIGFILENAME};

my $DOCBASEURI;
my $CHANNELMAP;
my $CONFIGROOTDIR;
my $CONFIGFILE;

##require '6ga-common.pl';
$configfile	= '/var/www/html/dvr/config/tv.conf';
readConfigFile($configfile);

# Hardcoded key for embedded action URI
my $faction	= "__6ga_form_action__";

################################
# Values coming from environ set by Apache
################################
my $clienturl	= $ENV{QUERY_STRING};
my $referer		= $ENV{HTTP_REFERER};

my $mycgi		= $ENV{SCRIPT_NAME};
my $reqmethod	= $ENV{REQUEST_METHOD};

while ($clienturl	=~ /%([[:xdigit:]]{2})/) {
	my $c	= chr(hex($1));
	$clienturl	=~ s/%$1/$c/ig;
}

################################
# Process POST method case
################################
if ($clienturl	=~ /$faction=([^&]+)/) {
	my $action	= $1;
	$clienturl	=~ s/$faction=$action//;
	$clienturl	=~ s/&&//;
	$clienturl	=~ s/^&//;
	$clienturl	=~ s/&$//;
	$clienturl	= "$action?$clienturl";
}
my $postdata;
if ($reqmethod eq "POST") {
	chomp($postdata	= <STDIN>);
}

my ($remote)	= split /\//, $clienturl, 1;
################################
# Send request and get response
################################
my ($content, @rheaders)  = wget("GET", $clienturl, $postdata);
##my ($content, @rheaders)  = w3m("GET", $clienturl, $postdata);

# Remove "HTTP/1.1" in response
shift @rheaders;

################################
# Scan headers
################################
my %headers;
my @headers;
while (my $hline = shift @rheaders) {
	my ($h, $v);
    if ($hline =~ /^\s+/) {
        $v  = ($_ =~ s/^\s+//);
        $headers{$h}    .= " $v";
    } else {
    	($h, $v)	= split /:\s+/, $hline, 2;
		if ($h eq "Location") {
			$clienturl	= trans_url($v);
			$v = "$mycgi?$clienturl";
			$hline = "$h: $v";
		}
    	$headers{$h}    = $v;
	}
	push @headers, $hline;
}

my $ctype	= $headers{'Content-Type'};
unless (defined($ctype)) {
	# tv.so-net.ne.jp uses 'Content-type'
	$ctype	= $headers{'Content-type'};
}

my $cdispos	= $headers{'Content-Disposition'};

if (defined($ctype) && $ctype =~ /^application\/x-tv-program-info/) {
	proc_iepg($content, $clienturl, @headers);
} elsif (defined($cdispos) && $cdispos =~ /attachment; filename=.*\.tvpi/) {
	# TV-Guide uses Content-Type: application/plain
	proc_iepg($content, $clienturl, @headers);
} elsif (defined($ctype) && $ctype =~ /^text\/html/) {
	if ($headers{'Content-Encoding'} || $headers{'Content-encoding'}) {
		proc_other($content, $clienturl, @headers);
	} else {
		proc_html($content, $clienturl, @headers);
	}
} else {
	proc_other($content, $clienturl, @headers);
}

exit 0;

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

########################################################
# iEPG file handler
########################################################
sub proc_iepg {
	my ($content, $iepgurl, @rheaders)	= @_;
	my %iepg;
	my ($iheader, $icontent)	= split /\n\n/, $content, 2;
	my $charset	= "shift_jis";	# Defaul code
	
	#########################
	# Parse MIME header
	#########################
	foreach (@rheaders) {
		my ($h, $v)	= split /:\s*/;

		# tv.so-net.ne.jp uses 'Content-type'
		if ($h =~ /^Content\-[Tt]ype$/
			&& defined($v)
			&& ($v =~ /;\s*charset=([\w\-]+)$/
				|| $v =~ /;\s*charset=["']([\w\-]+)["']$/)) {
			$charset	= $1;
		}
	}

	#########################
	# Parse header section
	#########################
	foreach (split /\n/, $iheader) {
		chomp;
		s/^\s+//;
		s/\s+$//;
		last if (/^$/);
		my ($h, $v)	= split /:\s+/;
		if (defined($h)) {
			# GET method has limit in URL length
			if (length $v > 40) {
				my $sppos	= index($v, ' ', 40);
				if ($sppos > 0) {
					$v	= substr $v, 0, $sppos;
				}
			}
			$iepg{$h}	= $v;
		}
	}

	$iepg{"channel"}	= findChannel($iepg{"station"});

	# Leave 'tab' undefined expecting the cookie will work
	my $PANELURI	= "$DOCBASEURI/6ga-panel.php?action=schedule";
	foreach my $key (keys %iepg) {
		$PANELURI	.= "&iepg_$key=" . uri_escape($iepg{$key});
	}
	$PANELURI	.= "&iepg_iepg-url=" . uri_escape($iepgurl);
	# Pass single quotes (apostrophes) in URI format
	$PANELURI	=~ s/\'/%27/g;

	my $dialogfile	= "dialog-$charset.js";

	print <<_END;
Content-Type: text/html; charset=$charset

<HTML><HEAD>
<SCRIPT LANGUAGE='JavaScript'>
<!--
	window.open('$PANELURI', '_6ga_panel');
	window.open('$referer', '_self');
-->
</SCRIPT>
</HAED><BODY>
</BODY></HTML>
_END
}

########################################################
# Look up station name for channel number
########################################################
sub findChannel {
	my ($station)	= @_;
	my $channel	= 0;

	open(CMAP, "$CONFIGROOTDIR/channels/$CHANNELMAP")
							|| die "can't open $CHANNELMAP";
	while (<CMAP>) {
		chomp;
		s/#.*$//;
		last if (/\[$station\]/);
	}

	while (<CMAP>) {
		chomp;
		s/#.*$//;
		if (/^\s*channel\s*=\s*(\d+)/) {
			$channel = $1;
			last;
		}
	}
	close CMAP;
	return $channel;
}

########################################################
# Convert URLs in HTML document
########################################################
sub proc_html {
	my ($rcontent, $iepgurl, @rheaders)	= @_;

	print join("\n", @rheaders), "\n\n";

	# Scan for tags
	my @rcontent = split /</, $rcontent;
	# Skip empty element ("1st" one)
	if (!$rcontent[0] || $rcontent[0] =~ /^\s*$/) {
		shift @rcontent;
	}

	my @content;
	my $incomment = 0;
	foreach my $cline (@rcontent) {
		if ($incomment) {
			$content[$#content]	.= "<$cline";
			$incomment = ($cline !~ /-->/s);
		} else {
			push @content, $cline;
			$incomment = ($cline =~ /^!--/s && $cline !~ /-->/s);
		}
	}

	my $inscript	= 0;
	foreach my $chunk (@content) {
		if ($chunk =~ /^!--/) {
			##############################
			# Comment block
			# Process JavaScript paths
			##############################
			if ($inscript) {
				$chunk =~ s/^(.*[\(,]\s*["'])(\/\w.*)$/$1$remote$2/gs;
			}
		} elsif ($chunk =~ /^(A[^>]*\sHREF\s*=\s*")([^"]+)(".*)$/is
           	|| $chunk =~ /^(A[^>]*\sHREF\s*=\s*')([^']+)('.*)$/is
           	|| $chunk =~ /^(A[^>]*\sHREF\s*=\s*)(\S+)(.*)$/is) {
			##############################
			# HTTP reference to another document - Put it under control
			##############################
			my ($x1, $x3)	= ($1, $3);
			my $turl	= trans_url($2);
			if ($turl =~ /^javascript:/i) {
				$chunk	= "$x1$turl$x3";
			} else {
				$chunk	= "$x1$mycgi?$turl$x3";
			}

			$chunk	=~ s/TARGET="_top"/TARGET="_6ga_main"/ig;
			$chunk	=~ s/TARGET='_top'/TARGET='_6ga_main'/ig;
		} elsif ($chunk =~ /^(LINK[^>]*\sHREF\s*=\s*")([^"]+)(".*)$/is
			|| $chunk =~ /^(LINK[^>]*\sHREF\s*=\s*')([^']+)('.*)$/is) {
			##############################
			# LINK REL="stylesheet" HREF="...
			##############################
			my ($x1, $x3)   = ($1, $3);
			my $turl    = trans_url($2);
			$chunk  = "$x1$turl$x3";
		} elsif ($chunk =~ /^(DIV \s*STYLE="[^"]+:url\(')([^'"]+'\)[^"]*".*)$/is
		|| $chunk =~ /^(DIV \s*STYLE='[^']+:url\(")([^'"]+"\)[^']*'.*)$/is) {
			##############################
			# DIV STYLE=background:url('...
			##############################
			my ($x1, $x2)   = ($1, $2);
			my $turl    = trans_url($2);
			$chunk  = "$x1$turl";
		} elsif ($chunk =~ /^(FORM[^>]*\sACTION\s*=\s*")([^"]*)(".*)$/is
           	|| $chunk =~ /^(FORM[^>]*\sACTION\s*=\s*')([^']*)('.*)$/is
           	|| $chunk =~ /^(FORM[^>]*\sACTION\s*=\s*)(\S*)(.*)$/is) {
			##############################
			# FORM action - Put it under control
			##############################
			my ($x1, $x3)	= ($1, $3);
			my $turl	= trans_url($2);
			$chunk	= "$x1$mycgi$x3<INPUT TYPE='HIDDEN' NAME='$faction' VALUE='$turl'>";

			$chunk	=~ s/TARGET="_top"/TARGET="_6ga_main"/ig;
			$chunk	=~ s/TARGET='_top'/TARGET='_6ga_main'/ig;

			############################
			# Add iEPG flag to yahoo!
			############################
			if ($turl =~ /^http:\/\/tv.yahoo.co.jp\//) {
				$chunk .= "<INPUT TYPE='HIDDEN' NAME='g' VALUE='2'>";
			}
		} elsif ($chunk =~ /^(META\s+HTTP-EQUIV\s*=\s*"Refresh"[^>]*\sCONTENT="\d+;\s*URL=)([^"]+)(".*)$/is
			|| $chunk =~ /^(META\s+HTTP-EQUIV\s*=\s*'Refresh'[^>]*\sCONTENT='\d+;\s*URL=)([^']+)('.*)$/is
			|| $chunk =~ /^(META\s+HTTP-EQUIV\s*=\s*Refresh[^>]*\sCONTENT=\d+;\s*URL=)([\S^>]+)(.*)$/is) {
			##############################
			# Refresh/jump meta header
			##############################
			my ($x1, $x3)	= ($1, $3);
			my $turl	= trans_url($2);
			$chunk	= "$x1$mycgi?$turl$x3";
		} else {
			if ($chunk =~ /^SCRIPT/is) {
				$inscript	= 1;
			}
			if ($chunk =~ /^\/SCRIPT/is) {
				$inscript	= 0;
			}
			##############################
			# Scan for other cases to fix relative URLs to absolute
			##############################
			foreach my $key ("SRC", "HREF", "BACKGROUND") {
				my $tchunk	= parse_url($chunk, $key);
				if (defined($tchunk)) {
					$chunk = $tchunk;
					last;
				}
			}
		}

		if ($chunk =~ /onclick="javascript:popup\('([^']+)'\)/) {
			##############################
			# TV-Guide's popup
			##############################
			my $path	= $1;
			my $tpath	= trans_url($path);
			$path	=~ s/\?/\\?/g;
			$path	=~ s/\./\\./g;
			$chunk	=~ s/$path/$mycgi?$tpath/g;
		}

		while ($chunk =~ /onmouseover="flip\(this,[ ]?'(img\/.*\.png)'\)"/) {
			##############################
			# The Television Javascript arg
			##############################
			my $path	= $1;
			my $tpath	= trans_url($path);
			$path	=~ s/\?/\\?/g;
			$path	=~ s/\./\\./g;
			$chunk	=~ s/$path/$tpath/g;
		}

		if ($chunk =~ /onchange=".+'([^']+\.php)/) {
			##############################
			# The Television Javascript arg
			##############################
			my $script	= $1;
			my $tscript	= trans_url($script);
			$script	=~ s/\./\\./g;
			$chunk	=~ s/$script/$tscript/g;
		}

		while ($chunk =~ /'([[:alnum:]]+\.php)['\?]/) {
			##############################
			# The Television Javascript arg
			##############################
			my $path	= $1;
			my $tpath	= trans_url($path);
			$path	=~ s/\?/\\?/g;
			$path	=~ s/\./\\./g;
			$chunk	=~ s/$path/$mycgi?$tpath/g;
		}

		while ($chunk =~ /'(http:\/\/[[:alnum:]\.\/]+\.php)/) {
			##############################
			# The Television Javascript arg
			##############################
			my $path	= $1;
			$chunk	=~ s/$path/$mycgi?$path/g;
			$chunk	=~ s/\.php\?/\.php%3f/g;
		}

		if ($chunk =~ /:url\(([^\)]+)\)/) {
			##############################
			# The Television Javascript arg
			##############################
			my $path	= $1;
			my $tpath	= trans_url($path);
			$path	=~ s/\?/\\?/g;
			$path	=~ s/\./\\./g;
			$chunk	=~ s/$path/$tpath/g;
		}

		print "<$chunk" if ($chunk && $chunk !~ /^\s*$/);
	}
}

########################################################
# Other document
########################################################
sub proc_other {
	my ($content, $iepgurl, @headers)	= @_;

	print join("\n", @headers), "\n\n";
	print $content;
}

########################################################
# Subroutine to convert relative URL to absolute in a tag
########################################################
sub parse_url {
	my ($tag, $key)	= @_;

	if ($tag =~ /^([^>]+\s${key}\s*=\s*")([^"]+)(".*)$/is
        || $tag =~ /^([^>]+\s${key}\s*=\s*')([^']+)('.*)$/is
        || $tag =~ /^([^>]+\s${key}\s*=\s*)(\S+)(.*)$/is) {
		my ($x1, $x3)	= ($1, $3);
		my $turl	= trans_url($2);
		if ($turl	=~ /^http:/) {
			return "$x1$turl$x3";
		}
	}
	return undef;
}

########################################################
# Subroutine to convert relative URL to absolute
########################################################
sub trans_url {
	my ($url)	= @_;
	my $curl	= $clienturl;
	my $rurl;

	$url	=~ s/^\s+//;
	$url	=~ s/\s+$//;
	if ($url =~ /^[[:alnum:]]+:\/\//) {
		# Full and absolute URL - pass it as-is
		$rurl = $url;
	} elsif ($url =~ /^javascript:[^\/]/i) {
		# Javascript - return it as-is
		$rurl = $url;
	} elsif ($url =~ /^\//) {
		# Absolute
		if ($curl =~ /^([[:alnum:]]+:\/\/[^\/]+)/) {
			$curl = $1;
		}
		$rurl = $curl . $url;
	} elsif ($url =~ /^\?/) {
		# Scrip parameters only - preced with script path
		if ($curl =~ /^([[:alnum:]]+:\/\/[^\?]+)/) {
			$curl = $1;
		}
		$rurl = $curl . $url;
	} else {
		$curl =~ s/([^\/]\/)[^\/]+$/$1/;
		if ($curl !~ /\/$/) {
			$curl .= "/";
		}
		$rurl = $curl . $url;
	}

	############################
	# Add iEPG flag to yahoo!
	############################
	if ($rurl =~ /^http:\/\/tv.yahoo.co.jp\/.*\.html$/) {
		$rurl .= "?g=2";
	}
	return $rurl;
}


########################################################
# Use w3m to get target page source and header
########################################################
sub w3m {
	my ($method, $url, $postdata)	= @_;
	my $tmpfile	= "/tmp/$myname.$$";
	my $wopt	= '';
	my @headers	= ();
	my $content	= "";

	if (defined($postdata)) {
		open TMP, ">$tmpfile"	|| die "Can't create $tmpfile\n";
		print TMP $postdata;
		close TMP;
		$wopt	= "--post-file=$tmpfile";
	}

	unless (open W3M, "HOME=/tmp w3m -config /dev/null -no-cookie -dump_both '$url'|") {
		unlink $tmpfile;
		die "Can't spawn 'w3m'\n";
	}
	while (<W3M>) {
		last if ($_ eq "\n");
		chomp;
		push @headers, $_;
	}

	while (<W3M>) {
		$content .= "$_";
	}
	close W3M;
	unlink $tmpfile;
	return ($content, @headers);
}

########################################################
# Local version of wget - obsolete
# Program wget has unstable options and header output format.
# Program wget also has a problem of silently redirecting to Location: URL
########################################################
sub wget {
	my ($method, $url, $postdata)	= @_;
	my ($port, $iaddr, $paddr, $proto);
	my $EOL = "\015\012";

	($url =~ /^http:\/\//)	|| die "Unsupported protocol in $url\n";
	$url =~ s/^http:\/\///;
	my ($remote, $uri)		= split /\//, $url, 2;
	if (!defined($uri)) {
		$uri = "";
	}
	($remote, $port)	= split /:/, $remote, 2;
	if (!defined($port)) {
		$port	= 80;
	}

	if ($port =~ /\D/) {
		$port = getservbyname($port, 'tcp')
	}

	$iaddr   = gethostbyname($remote)			|| die "no host: $remote";
	$paddr   = sockaddr_in($port, $iaddr);
	$proto   = getprotobyname('tcp');
	socket(SOCK, PF_INET, SOCK_STREAM, $proto)	|| die "socket: $!";
	connect(SOCK, $paddr)						|| die "connect: $!";

	SOCK->autoflush(1);
	my $accept		= $ENV{HTTP_ACCEPT};
	my $user_agent	= $ENV{HTTP_USER_AGENT};
	print SOCK "$method /$uri HTTP/1.0$EOL"
				. "User-Agent: $user_agent$EOL"
				. "Accept: $accept$EOL"
				. "Host: $remote$EOL";

	# Get cookies
	my $cookie	= $ENV{HTTP_COOKIE};
	my @cookies	= split(/ /, $cookie) if (defined($cookie));

	foreach my $ck (@cookies) {
		print SOCK "Cookie: $ck$EOL";
	}
	print SOCK "$EOL";
	print SOCK "$postdata$EOL"	if (defined($postdata));

	my @headers	= ();
	my $content	= "";
	while (<SOCK>) {
		last if ($_ eq "\n" || $_ eq $EOL);
		chomp;
		push @headers, $_;
	}
	while (<SOCK>) {
		chomp; chomp;
		$content .= "$_\n";
	}

	##close SOCK            || die "close: $!";
	close SOCK;
	return ($content, @headers);
}

########################################################
# Read in cofiguration file
########################################################
sub readConfigFile {
	my ($configfile)	= @_;
	(defined $configfile)	|| die "no config file\n";

	open(CONFIG, $configfile)	|| die "config file $configfile not found/\n";
	while (<CONFIG>) {
		chomp;
		eval;
	}
	close CONFIG;
}

